
/*

source code components :
  pppoem.c pppoem.h

purpose :
  Monitor the DSL connection of a Linux 2.4.4 system (e.g. SuSE 7.2)
  See also help text in main() and output of: cat /proc/net/pppoe /proc/net/dev

compilation :
  cc -g -DPppoem_use_own_sfilE -DPppoem_use_own_maiN -o pppoem pppoem.c

usage :
  after compilation just start  ./pppoem  in a terminal window
  start  ./pppoem -help  to get the helptext on stderr.

license and support address :
  BSD license, copyright 2001 Thomas Schmitt, scdbackup@gmx.net
  use it as you like

*/


#ifndef Pppoem_no_system_includeS

#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <time.h>
#include <wait.h>
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>
#include <signal.h>

#endif /* ! Pppoem_no_system_includeS */

/* ------------------------------------------------------------------------ */


#ifdef Pppoem_use_own_maiN


#define Pppoem_use_own_cleanuP

static struct PppoeM *PppoemM= NULL;


main(argc,argv)
int argc;
char **argv;
{
 char dev_name[80];
 int ret;

 /*
    Define the default line to watch in /proc/net/dev .
    dev_name "-" tries to read the name from
    the second line of /proc/net/pppoe (like "eth0")
    as soon as it gets available.
    One could prefer to watch "ppp0" though. Anyway
    argument -dev_name is able to override the default.
    loop delay 1.0 second, report intervals 1, 15, 300
    standalone | local_time | (only one headline)
 */
 strcpy(dev_name,"-");
 ret= Pppoem_new(&PppoemM,dev_name,1.0,1,15,300,1|2);
 if(ret<=0)
   exit(1);

 ret= Pppoem_arg_interpreter(PppoemM,argc,argv,1);
 if(ret<=0)
   exit(4);
 if(argc==2 && (strcmp(argv[1],"-help")==0 || strcmp(argv[1],"-h")==0)) 
   exit(0);

 /* to get a message on exit */
 Cleanup_set_handlers(0);

 ret= Pppoem_loop(PppoemM,0);
 if(ret<=0)
   exit(2);

 Pppoem_report(PppoemM,"PPPOEM E",!Pppoem_is_open(PppoemM,0));
 
 exit(0);
}

#endif /* Pppoem_use_own_maiN */


#ifdef Pppoem_use_own_cleanuP

/* !!! needs a global  struct PppoeM *PppoemM  */

void Cleanup_handler_exit()
{
 if(PppoemM!=NULL)
   Pppoem_report(PppoemM,"PPPOEM X",!Pppoem_is_open(PppoemM,0));
 exit(3);
} 


int Cleanup_set_handlers(flag)
int flag;
/*
 bit0= set to default handlers
*/
{
 if(flag&1)
   goto set_default;

 signal(SIGHUP,Cleanup_handler_exit);
 signal(SIGINT,Cleanup_handler_exit);
 signal(SIGQUIT,Cleanup_handler_exit);
 signal(SIGILL,Cleanup_handler_exit);
 signal(SIGTRAP,Cleanup_handler_exit);
 signal(SIGFPE,Cleanup_handler_exit);
 signal(SIGBUS,Cleanup_handler_exit);
 signal(SIGSEGV,Cleanup_handler_exit);
 signal(SIGPIPE,Cleanup_handler_exit);
 signal(SIGALRM,Cleanup_handler_exit);
 signal(SIGTERM,Cleanup_handler_exit);
 signal(SIGTERM,Cleanup_handler_exit);
 signal(SIGUSR1,Cleanup_handler_exit);
 signal(SIGUSR2,Cleanup_handler_exit);
 signal(SIGURG,Cleanup_handler_exit);
 signal(SIGXCPU,Cleanup_handler_exit);
 signal(SIGXFSZ,Cleanup_handler_exit);
 signal(SIGVTALRM,Cleanup_handler_exit);
 signal(SIGPROF,Cleanup_handler_exit);
 return(1);
set_default:;
 signal(SIGHUP,SIG_DFL);
 signal(SIGINT,SIG_DFL);
 signal(SIGQUIT,SIG_DFL);
 signal(SIGILL,SIG_DFL);
 signal(SIGTRAP,SIG_DFL);
 signal(SIGFPE,SIG_DFL);
 signal(SIGBUS,SIG_DFL);
 signal(SIGSEGV,SIG_DFL);
 signal(SIGPIPE,SIG_DFL);
 signal(SIGALRM,SIG_DFL);
 signal(SIGTERM,SIG_DFL);
 signal(SIGUSR1,SIG_DFL);
 signal(SIGUSR2,SIG_DFL);
 signal(SIGURG,SIG_DFL);
 signal(SIGXCPU,SIG_DFL);
 signal(SIGXFSZ,SIG_DFL);
 signal(SIGVTALRM,SIG_DFL);
 signal(SIGPROF,SIG_DFL);
 return(1);
}

#endif /* Pppoem_use_own_cleanuP */


#ifdef Pppoem_use_smeM

#include "../s_tools/smem.h"

#else /* Pppoem_use_smeM */

#define TSOB_FELD(typ,anz) (typ *) malloc((anz)*sizeof(typ));
#define Smem_malloC malloc
#define Smem_freE free

#endif /* ! Pppoem_use_smeM */


#ifdef Pppoem_use_own_sfilE

/* ----------------------------------  SfilE  ---------------------------- */

int Sfile_microsleep(microsec,flag)
/* sleep very fast by watching no file */
unsigned long microsec;
int flag;
{
 struct timeval wt;
 fd_set rds,wts,exs;
 int ready;

 wt.tv_sec=  microsec/1000000;
 wt.tv_usec= microsec%1000000;
 FD_ZERO(&rds);
 FD_ZERO(&wts);
 FD_ZERO(&exs);
 ready= select(1,&rds,&wts,&exs,&wt);
 if(ready>0)
   return(0); /* should never happen */
 return(1);
}


double Sfile_microtime(flag)
int flag;
{
 struct timeval tv;
 struct timezone tz;
 gettimeofday(&tv,&tz);
 return((double) (tv.tv_sec+1.0e-6*tv.tv_usec));
}


char *Sfile_datestr(tim,flag)
time_t tim;
short int flag;
/*
 bit0=with hours+minutes  
 bit1=with seconds

 bit8= local time rather than UTC
*/
{
 static char zeitcode[80]={"000000"};
 char puff[80];
 struct tm *azt;

 if(flag&256) 
   azt = localtime(&tim);
 else
   azt = gmtime(&tim);
 
 if(azt->tm_year>99)
   sprintf(zeitcode,"%c%01.1d%02.2d%02.2d",
           'A'+(azt->tm_year-100)/10,azt->tm_year%10,
           azt->tm_mon+1,azt->tm_mday);
 else
   sprintf(zeitcode,"%02.2d%02.2d%02.2d",
           azt->tm_year,azt->tm_mon+1,azt->tm_mday);
 if(flag&1){
   sprintf(puff,".%02.2d%02.2d",azt->tm_hour,azt->tm_min);
   strcat(zeitcode,puff);
 }
 if(flag&2){
   sprintf(puff,"%02.2d",azt->tm_sec);
   strcat(zeitcode,puff);
 }

 return(zeitcode);
}


int Sfile_scale(value,result,siz,thresh,flag)
double value;
char *result;
int siz;
double thresh;
int flag;
/*
 bit0= eventually ommit 'b'
 bit1= make text as short as possible
 bit2= no fraction (if it would fit at all)
*/
{
 char scale_c,scales[6],form[80];
 int i,dec_siz= 0,avail_siz;

 strcpy(scales,"bkmgt");
 scale_c= scales[0];
 for(i=1;scales[i]!=0;i++) {
   if(value<thresh) 
 break;
   value/= 1024;
   scale_c= scales[i];
 } 
 if(scale_c!='b' && !(flag&4)) { /* is there room for fractional part ? */
   avail_siz= siz-1;
   sprintf(form,"%%.lf");
   sprintf(result,"%.lf",value);
   if(strlen(result)<avail_siz-2) 
     dec_siz= 1;                                  /* we are very modest */
 }
 if(scale_c=='b' && (flag&1)) {
   if(flag&2)
     sprintf(form,"%%.lf");
   else
     sprintf(form,"%%%ld.lf",siz);
   sprintf(result,form,value);
 } else {
   if(flag&2)
     sprintf(form,"%%.lf%%c");
   else if(dec_siz>0) 
     sprintf(form,"%%%ld.%ldlf%%c",avail_siz,dec_siz);
   else
     sprintf(form,"%%%ld.lf%%c",siz-1);
   sprintf(result,form,value,scale_c);
 }
 return(1);
}


int Sfile_fd_is_ready(fd,microsec,flag)
int fd;
unsigned int microsec;
int flag;
/*
 bit0= check for reading
 bit1= check for writing
 bit2= check for exception
return-bits are set according to matching flag-bits
*/
{
 struct timeval wt;
 fd_set rds,wts,exs;
 int ready,ret;

 wt.tv_sec=  microsec/1000000;
 wt.tv_usec= microsec%1000000;
 FD_ZERO(&rds);
 FD_ZERO(&wts);
 FD_ZERO(&exs);
 if(flag&1)
   FD_SET(fd,&rds);
 if(flag&2)
   FD_SET(fd,&wts);
 if(flag&4)
   FD_SET(fd,&exs);
 ready= select(fd+1,&rds,&wts,&exs,&wt);
 if(ready<=0)
   return(0);
 ret= 0;
 if(flag&1)
   if(FD_ISSET(fd,&rds))
     ret|=1;
 if(flag&2)
   if(FD_ISSET(fd,&wts))
     ret|=2;
 if(flag&4)
   if(FD_ISSET(fd,&exs))
     ret|=4;
 return(ret);
}

#endif /* Pppoem_use_own_sfilE */


/* ----------------------------------  PppoeM  ---------------------------- */


#include "pppoem.h"


int Pppoem_new(mon,dev_name,delay,span1,span2,span3,flag)
struct PppoeM **mon;
char *dev_name;
double delay;
int span1,span2,span3;
int flag;
/*
 bit0= standalone
 bit1= use local time for time stamps
 bit2= print headlines to stderr each time a connection starts
*/
{
 struct PppoeM *m;
 int i,ret;
 
 *mon= m= TSOB_FELD(struct PppoeM,1);
 if(m==NULL)
   return(-1);

 m->standalone= !(flag&1);
 m->listen_stdin= 0;
 m->dev_name[0]= 0;
 m->dev_automatic= 1;
 m->delay= delay;
 m->count_what= 3;

 m->time_spans[0]= span1;
 m->time_spans[1]= span2;
 m->time_spans[2]= span3;
 m->quiet= 0; 
 m->use_local_time= !!(flag&2);
 m->headlines_on= !!(flag&4);
 m->deadline= 0;

 m->poll_count= 0;
 m->conn_count= 0;
 m->missing_pppoe_silent= 0;
 m->missing_pppoe_reported= 0;
 m->is_open= -1;
 m->start_time= 0;
 m->start_bytes= 0;
 m->last_time= 0;
 m->last_bytes= 0;
 m->ring_counts= NULL;
 m->ring_times= NULL;
 m->ring_size= span1+1;
 if(span2+1>m->ring_size)
   m->ring_size= span2+1;
 if(span3+1>m->ring_size)
   m->ring_size= span3+1;
 if(m->ring_size<10)
   m->ring_size= 10;
 m->ring_cursor= 0;
 m->ring_full= 0;

 if(strlen(dev_name)>=sizeof(m->dev_name))
   goto failed;
 strcpy(m->dev_name,dev_name);
 if(dev_name[0]!=0 && strcmp(dev_name,"-")!=0)
   m->dev_automatic= 0;
 ret= Pppoem_new_ring(m,m->ring_size,0);
 if(ret<=0)
   goto failed;

/*
 m->ring_counts= TSOB_FELD(double,m->ring_size);
 if(m->ring_counts==NULL)
   goto failed;
 m->ring_times= TSOB_FELD(double,m->ring_size);
 if(m->ring_times==NULL)
   goto failed;
 for(i=0;i<m->ring_size;i++)
   m->ring_counts[i]= m->ring_times[i]= 0.0;
*/

 return(1);

failed:;
 Pppoem_destroy(mon,0);
 return(-1);
}


int Pppoem_new_ring(mon,ring_size,flag)
struct PppoeM *mon;
int ring_size;
int flag;
{
 double *ring_counts= NULL,*ring_times= NULL;
 int ret,ring_cursor= 0,ring_full= 0,i,idx;

 if(mon->ring_counts!=NULL && mon->ring_times!=NULL
    && mon->ring_size==ring_size)
   return(2);

 ring_counts= TSOB_FELD(double,ring_size);
 if(ring_counts==NULL)
   {ret= -1; goto failed;}
 ring_times= TSOB_FELD(double,ring_size);
 if(ring_times==NULL)
   {ret= -1; goto failed;}
 for(i=0;i<ring_size;i++)
   ring_counts[i]= ring_times[i]= 0.0;
 if(mon->ring_counts!=NULL && mon->ring_times!=NULL) {
   if(mon->ring_size>ring_size) {
     if(mon->ring_full || mon->ring_cursor>=ring_size) {
       for(i= -ring_size;i<0;i++) {
         Pppoem_ring_index(mon,mon->ring_cursor+i,&idx,0);
         ring_counts[ring_size+i]= mon->ring_counts[idx];
         ring_times[ring_size+i]= mon->ring_times[idx];
       }
       ring_cursor= 0;
       ring_full= 1;
     } else {
flat_copy:;
       for(i=0;i<mon->ring_cursor;i++) {
         ring_counts[i]= mon->ring_counts[i];
         ring_times[i]= mon->ring_times[i];
       }
       ring_cursor= mon->ring_cursor;
       ring_full= 0;
     }
   } else {
     if(mon->ring_full) {
       for(i=0;i<mon->ring_size;i++) {
         Pppoem_ring_index(mon,mon->ring_cursor+i,&idx,0);
         ring_counts[i]= mon->ring_counts[idx];
         ring_times[i]= mon->ring_times[idx];
       }
       ring_cursor= mon->ring_size;
       ring_full= 0;
     } else {
       goto flat_copy;
     }
   }
   Smem_freE((char *) mon->ring_counts);
   mon->ring_counts= NULL;
   Smem_freE((char *) mon->ring_times);
   mon->ring_times= NULL;
 }
 mon->ring_size= ring_size;
 mon->ring_counts= ring_counts;
 mon->ring_times= ring_times;
 mon->ring_cursor= ring_cursor;
 mon->ring_full= ring_full;
 return(1);

failed:;
 if(ring_counts!=NULL)
   Smem_freE((char *) ring_counts);
 if(ring_times!=NULL)
   Smem_freE((char *) ring_times);
 return(ret);
}


int Pppoem_destroy(mon,flag)
struct PppoeM **mon;
int flag;
{
 struct PppoeM *m;

 m= *mon;
 if(m==NULL)
   return(0);
 if(m->ring_counts!=NULL)
   Smem_freE((char *) m->ring_counts);
 m->ring_counts= NULL;
 if(m->ring_times!=NULL)
   Smem_freE((char *) m->ring_times);
 m->ring_times= NULL;
 Smem_freE((char *) m);
 *mon= NULL;
}


int Pppoem_is_open(mon,flag)
struct PppoeM *mon;
int flag;
{
 if(mon==NULL)
   return(0);
 return(mon->is_open);
}


int Pppoem_has_headlines_on(mon,flag)
struct PppoeM *mon;
int flag;
{
 if(mon==NULL)
   return(0);
 return(mon->headlines_on);
}


int Pppoem_has_lifetime(mon,flag)
struct PppoeM *mon;
int flag;
{
 time_t t;

 if(mon->deadline==0)
   return(1<<30);
 t= time(0);
 if(t<=mon->deadline)
   return(mon->deadline-t);
 return(0);
}


int Pppoem__help_stderr(flag)
int flag;
/*
 bit0= also print pppoem options (see Pppoem_arg_interpreter)
*/
{
 fprintf(stderr,
    "This program monitors the DSL connection of a Linux kernel 2.4\n");
 fprintf(stderr,
    "by reading the files /proc/net/pppoe and /proc/net/dev once a second.\n");
 fprintf(stderr,
    "It prints a status report on stderr once per second and a change\n");
 fprintf(stderr,
    "notification on stdout each time the connection changes its state.\n");
 fprintf(stderr,
    "The status report is a text line containing the fields :\n");
 fprintf(stderr,
    "'State' which is either 'DSL OFF' or 'DSL ON' or a device name\n");
 fprintf(stderr,
    "'Time (local)' which is a timestamp like 'YYMMDD.hhmmss'\n");
 fprintf(stderr,
    "'1s' , '15s' , '300s' show the average bytes-per-second for the last\n");
 fprintf(stderr,
    " 1, 15 resp. 300 samples (resp. -1 until valid).\n");
 fprintf(stderr,
    "'total =  bytes / uptime' shows a summary since the connection went ON:\n"
        );
 fprintf(stderr,
    " bytes-per-second = sent plus received bytes / life time in seconds\n");
 fprintf(stderr,
    "Snapshot of Tue Oct 23 16:08:18 CEST 2001 :\n");
 fprintf(stderr,
   " State    Time (local)      1s     15s    300s   total =  bytes / uptime\n"
        );
 fprintf(stderr,
    "DSL ON   A11023.160818   53410   37265      -1   22145 =  5712k / 264s\n"
        );
 fprintf(stderr,
    "\n");

 if(!(flag&1))
   return(1);

 fprintf(stderr,
    "Options of pppoem :\n");
 fprintf(stderr,
    "  -count received|sent|all\n");
 fprintf(stderr,
    "         If set to \"received\" only the \"Receive\" side of the\n");
 fprintf(stderr,
    "         device in /proc/net/dev will be watched. If set to \"sent\"\n");
 fprintf(stderr,
    "         the same is done for \"Transmit\" only. If set to \"all\"\n");
 fprintf(stderr,
    "         both sides will count (this is the default).\n");
 fprintf(stderr,
    "  -dev_name name\n");
 fprintf(stderr,
    "         Set the name of the device line in /proc/net/dev which shall\n");
 fprintf(stderr,
    "         be watched to obtain throughput information. If name is given\n");
 fprintf(stderr,
    "         as \"-\" then it will be read from /proc/net/pppoe when the\n");
 fprintf(stderr,
    "         connection is in state ON (this is the default setting).\n");
 fprintf(stderr,
    "  -pretend on|off\n");
 fprintf(stderr,
    "         If set to \"on\" ignore /proc/net/pppoe and always watch in\n");
 fprintf(stderr,
    "         /proc/net/dev the device given with -dev_name . The state is\n");
 fprintf(stderr,
    "         like permanent ON but the -dev_name gets displayed instead.\n");
 fprintf(stderr,
    "         One may monitor non-pppoe net devices like \"lo\" this way.\n");
 fprintf(stderr,
    "         It might even work with older kernels.\n");
 fprintf(stderr,
    "  -lifetime seconds\n");
 fprintf(stderr,
    "         Causes monitoring to end after the given time span. 0=endless\n");
 fprintf(stderr,
    "  -listen_stdin on|off\n");
 fprintf(stderr,
    "         If set to \"on\" ignore /proc . Listen for messages at stdin:\n");
 fprintf(stderr,
    "         \"%%ld %%s %%lf %%lf\" = is_on, dev_name, timestamp, count\n");
 fprintf(stderr,
    "         (is_on: 0=OFF,1=ON,2=-pretend, timestamp: seconds since 1970)\n");
 fprintf(stderr,
    "  -report  on|off|forward\n");
 fprintf(stderr,
    "         Control reporting on stdout and stderr. \"off\" disables all\n");
 fprintf(stderr,
    "         regular messages. \"on\" is the normal mode. \"forward\"\n"); 
 fprintf(stderr, 
    "         prints messages to stdout suitable for -listen_stdin \"on\".\n");
 fprintf(stderr,
    "  -show  number:number:number\n");
 fprintf(stderr,
    "         Set the numbers of samples for the three gliding averages\n");
 fprintf(stderr,
    "         shown in the status line. Default is 1:15:300 .\n");
 fprintf(stderr,
    "  -wait_silently on|off\n");
 fprintf(stderr,
    "         If on: do not emit error messages about missing /proc/net/pppoe\n");
 fprintf(stderr,
    "  -help\n");
 fprintf(stderr,
    "         Print this text to stderr.\n");
 fprintf(stderr,
    "\n");
 return(1);
}


int Pppoem_poll_open(mon,flag)
struct PppoeM *mon;
int flag;
{
 int started= 0,linecount=0,ret= 0;
 char line[4096],filename[80],dummy[4096];
 FILE *fp= NULL;

 if(mon->is_open==2) {
   if(mon->dev_name[0]==0 || strcmp(mon->dev_name,"-")==0 ||
      mon->dev_automatic) {
     fprintf(stderr,"-pretend \"on\" needs -dev_name\n");
     return(-1);
   } else
     return(2);
 }

 strcpy(filename,"/proc/net/pppoe");
 fp= fopen(filename,"r");
 if(fp==NULL) {
   if(!(mon->missing_pppoe_reported || mon->missing_pppoe_silent)) {
     sprintf(line,"\ncannot open %s ",filename);
     perror(line);
     mon->missing_pppoe_reported= 1;
   }
   ret= 0; goto ex;
 } else {
   if(mon->missing_pppoe_reported)
     fprintf(stderr,"previously missing file %s appeared\n",filename);
   mon->missing_pppoe_reported= 0;
 }
 while(!feof(fp)) {
   if(fgets(line,sizeof(line),fp)==NULL)
 break;
   linecount++;
   if(mon->dev_name[0]==0 || strcmp(mon->dev_name,"-")==0 || mon->dev_automatic)
     if(linecount==2)
       sscanf(line,"%s %s %s",dummy,dummy,mon->dev_name);
 }
 mon->is_open= (linecount>1);
 ret= mon->is_open;
ex:;
 if(fp!=NULL)
   fclose(fp);
 return(ret);
}


int Pppoem_poll_count(mon,flag)
struct PppoeM *mon;
int flag;
{
 char line[1001],filename[80],*cpt;
 FILE *fp= NULL;
 int ret,i;
 double received,sent,r,bytes,diff;
 
 mon->ring_counts[mon->ring_cursor]= 0;
 mon->ring_times[mon->ring_cursor]= Sfile_microtime(0);

 if(mon->dev_name[0]==0 || strcmp(mon->dev_name,"-")==0)
   {ret= 2; goto ex;} 

 /* read in /proc/net/dev from line $dev_name column 0 and 8 */

 strcpy(filename,"/proc/net/dev");
 fp= fopen(filename,"r");
 if(fp==NULL) {
   sprintf(line,"cannot open %s ",filename);
   perror(line);
   ret= -1; goto ex;
 }
 while(!feof(fp)) {
   if(fgets(line,sizeof(line),fp)==NULL)
 break;
   for(cpt=line;(*cpt)!=0;cpt++)
     if((*cpt)!=' ')
   break;
   if(strncmp(cpt,mon->dev_name,strlen(mon->dev_name))!=0)
 continue;
   cpt+= strlen(mon->dev_name);
   if((*cpt)!=':')
 continue;
   cpt++;
   sscanf(cpt,"%lf  %lf  %lf  %lf  %lf  %lf  %lf  %lf  %lf",
          &received,&r  ,&r  ,&r  ,&r  ,&r  ,&r  ,&r  ,&sent);
   bytes= 0;
   if(mon->count_what&1)
     bytes+= received;
   if(mon->count_what&2)
     bytes+= sent;
 
   if(mon->ring_cursor>0 || mon->ring_full) {
     diff= bytes - mon->last_bytes;

     /* >>> check for plausibility */
     while(diff<0) {
       diff+= 4294967296.0;
       mon->start_bytes-= 4294967296.0;
     }

     mon->ring_counts[mon->ring_cursor]= diff;
   }
   mon->last_time= mon->ring_times[mon->ring_cursor]= Sfile_microtime(0);
   mon->last_bytes= bytes;
   if(mon->start_time==0.0) {
     mon->start_time= mon->last_time;
     mon->start_bytes= mon->last_bytes;
   }
   ret= 1; goto ex;
 }
 ret= 0;
ex:;
 Pppoem_ring_index(mon,mon->ring_cursor+1,&(mon->ring_cursor),1);
 if(mon->ring_cursor==0)
   mon->ring_full= 1;
 if(fp!=NULL)
   fclose(fp);
 return(ret);
}


int Pppoem_poll_stdin(mon,flag)
/*
poll for a "%ld %s %lf %lf" formatted message at stdin :
  mon->is_open , mon->dev_name , mon->ring_times[] , mon->ring_counts[]
*/
struct PppoeM *mon;
int flag;
/*
 bit0= read only a single line
*/
{
 int ret,is_open= -1,seen_open= 0,first_try= 1;
 char line[1024],dev_name[1024];
 double timestamp= -1.0, count= -1.0, total_count= 0.0;

 mon->ring_counts[mon->ring_cursor]= 0;
 mon->ring_times[mon->ring_cursor]= Sfile_microtime(0);

get_line:;
 /* read message from stdin */;
 if(fgets(line,sizeof(line)-1,stdin)==NULL) {
   perror("while reading from stdandard input :");
   return(-1);
 }

 /* decode */
 sscanf(line,"%ld %s %lf %lf",&is_open,dev_name,&timestamp,&count);
 if(timestamp<0 || count<0 )
   {ret= 0; goto ex;}

 /* >>> check for plausibility */

 total_count+= count;
 /* if(first_try) fprintf(stderr,"\n");
    fprintf(stderr,"total_count=%lf , count=%lf\n",total_count,count);
 */
 if(is_open>seen_open)
   seen_open= is_open;

 /* to compensate interfering intervals one must be able to
    read more than one sample per try */
 first_try= 0;
 if(!(flag&1) && Sfile_fd_is_ready(fileno(stdin),0,1)>0) {
   /* fprintf(stderr,"extra line at stdin detected\n"); */
   goto get_line;
 }

 mon->is_open= seen_open;
 strncpy(mon->dev_name,dev_name,sizeof(mon->dev_name)-1);
 mon->dev_name[sizeof(mon->dev_name)-1]= 0;
 mon->last_time= mon->ring_times[mon->ring_cursor]= timestamp;
 mon->ring_counts[mon->ring_cursor]= total_count;
 mon->last_bytes+= total_count;

 if(mon->start_time==0.0) {
   mon->start_time= mon->last_time;
   mon->start_bytes= mon->last_bytes;
 }
 ret= 1;
ex:;
 Pppoem_ring_index(mon,mon->ring_cursor+1,&(mon->ring_cursor),1);
 if(mon->ring_cursor==0)
   mon->ring_full= 1;
 return(ret);
}


int Pppoem_poll(mon,flag)
struct PppoeM *mon;
int flag;
/*
 bit0= in listen_stdin-mode: one line only
*/
{
 int ret,was_open,idx;

 if(mon==NULL)
   return(0);

 was_open= mon->is_open;
 if(mon->listen_stdin) {
   ret= Pppoem_poll_stdin(mon,flag&1);
   if(ret<0)
     goto ex;
 } else {
   ret= Pppoem_poll_open(mon,0);
   if(ret<0)
     goto ex;
   ret= Pppoem_poll_count(mon,0);
   if(ret<0)
     goto ex;
 }
 mon->poll_count++;
 if(mon->is_open>0 && was_open<=0) {
   /* New connection */
   mon->start_time= mon->last_time;
   mon->start_bytes= mon->last_bytes;
   if(Pppoem_ring_index(mon,mon->ring_cursor-1,&idx,0)>0) {
     mon->ring_counts[0]= mon->ring_counts[idx];
     mon->ring_times[0]= mon->ring_times[idx];
   }
   mon->ring_cursor= 1;
   mon->ring_full= 0;
   mon->conn_count++;
 }
 ret= 1;
ex:;
 return(ret);
}


int Pppoem_ring_index(mon,idx,result,flag)
struct PppoeM *mon;
int idx;
int *result;
int flag;
/*
 bit0= allow unused indice
*/
{
 *result= 0;
 if((!mon->ring_full)&&!(flag&1)) 
   if(idx<0 || idx>=mon->ring_cursor)
     return(0);
 *result= idx % mon->ring_size;
 if(*result<0)
   (*result)+= mon->ring_size;
 return(1);
}


int Pppoem_get_interval(mon,off1,off2,count,timespan,flag)
struct PppoeM *mon;
int off1,off2;
double *count;
double *timespan;
int flag;
/*
 bit0= get whole interval of connection
*/
{
 int i,i1,i2,tr,ret;

 *count= 0;
 *timespan= 0;

 if(mon==NULL)
   return(0);

 if(flag&1) {
   *timespan= mon->last_time - mon->start_time;
   *count= mon->last_bytes - mon->start_bytes;
   return(1); 
 }
 if(off1>off2)
   { tr= off1; off1= off2; off2= tr;}
 if(off1==off2)
   off2++;
 if(off2>=mon->ring_size)
   return(0);
 ret= Pppoem_ring_index(mon,mon->ring_cursor-1-off1,&i1,0);
 if(ret<=0)
   return(ret);
 ret= Pppoem_ring_index(mon,mon->ring_cursor-1-off2,&i2,0);
 if(ret<=0)
   return(ret);
 for(i= i1; i!=i2; ) {
   (*count)+= mon->ring_counts[i];
   Pppoem_ring_index(mon,i-1,&i,0);
 }
 *timespan= mon->ring_times[i1]-mon->ring_times[i2];
 return(1);
}


int Pppoem_format_bps(mon,off1,off2,result,flag)
struct PppoeM *mon;
char result[160];
int flag;
{
 int ret;
 double count,timespan;

 ret= Pppoem_get_interval(mon,off1,off2,&count,&timespan,0);
 if(ret<0)
   return(ret);
 if(ret>0 && timespan>0) {
   Sfile_scale(count/timespan,result,7,1e5,1);
 } else
   sprintf(result,"%7.7s","     -1");
 return(1);
}


int Pppoem_format_stat(mon,result,flag)
struct PppoeM *mon;
char result[160];
int flag;
/*
 bit0= no on_off status
 bit1= no date string
 bit2= no throughput
 bit3= short line: No "DSL", only total throughput
*/
{
 int i,ret,idx;
 time_t tim;
 char bps1[80],bps2[80],bps3[80],total[80],on_off[80],datestr[80];
 char rate_text[80],count_text[80];
 char scale_c,scales[6];
 double t,count,rate;

 strcpy(scales,"bkmgt");
 result[0]= 0;

 if(mon==NULL)
   return(0);

 if(!(flag&1)) {
   if(mon->is_open==2)
     strncpy(on_off,mon->dev_name,sizeof(on_off)-1);
   else if(mon->is_open>0)
     sprintf(on_off,"DSL ON");
   else
     sprintf(on_off,"DSL OFF");
   on_off[sizeof(on_off)-1]= 0;
   if((flag&8) && mon->is_open==2)
     sprintf(result,"%.8s ",on_off);
   else if(flag&8) 
     sprintf(result,"%-3.3s ",on_off+4);
   else
     sprintf(result,"%-8.8s ",on_off);
 }
 if(!(flag&2)) {
   ret= Pppoem_ring_index(mon,mon->ring_cursor-1,&idx,0);
   if(ret>0)
     tim= mon->ring_times[idx];
   else
     tim= time(0);
   strcpy(datestr,Sfile_datestr(tim,3|(mon->use_local_time<<8)));
   sprintf(result+strlen(result),"%s ",datestr);
 }
 if(!(flag&(4|8))) {
   ret= Pppoem_format_bps(mon,0,mon->time_spans[0],bps1,0);
   if(ret<=0)
     return(ret);
   ret= Pppoem_format_bps(mon,0,mon->time_spans[1],bps2,0);
   if(ret<=0)
     return(ret);
   ret= Pppoem_format_bps(mon,0,mon->time_spans[2],bps3,0);
   if(ret<=0)
     return(ret);
   sprintf(result+strlen(result),"%s %s %s ",bps1,bps2,bps3);
 }
 if(!(flag&4)) {
   t= mon->last_time - mon->start_time;
   count= mon->last_bytes - mon->start_bytes;
   if(t>0)
     rate= count/t;
   else
     rate= 0;
   Sfile_scale(count,count_text,7,1e5,0);
   Sfile_scale(rate,rate_text,7,1e5,1);
   sprintf(total,"%s =%s / %.lfs",rate_text,count_text,t);
   strcat(result,total);
 }
 return(1);
}


int Pppoem_show_stat(mon,flag)
struct PppoeM *mon;
int flag;
/*
 bit0= clear status line and return
 bit1= print headline and return
*/
{
 char line[160];
 int ret;

 if(mon!=NULL)
   if(mon->quiet)
     return(2);

 fprintf(stderr,"\r%-79.79s\r"," ");
 fflush(stderr);
 if(flag&1)
   return(1);
 if(mon==NULL)
   return(0);

 if(flag&2)
   goto headline;

 ret= Pppoem_format_stat(mon,line,(mon->is_open<=0)<<2);
 if(ret<=0)
   return(ret);
 fprintf(stderr,"\r%s ",line);
 fflush(stderr);
 return(1);

headline:;
 fprintf(stderr,"%8s %13s %6lds %6lds %6lds  %s\n",
         " State  ",(mon->use_local_time?" Time (local)":" Time (GMT)"),
         (int) (mon->delay*mon->time_spans[0]),
         (int) (mon->delay*mon->time_spans[1]),
         (int) (mon->delay*mon->time_spans[2]),
         " total =  bytes / uptime");
 fflush(stderr);
 return(1);
}


int Pppoem_report(mon,msg,flag)
struct PppoeM *mon;
char *msg;
int flag;
/*
 bit0= do not print throughput information
 bit1= append "$dev_name"
*/
{
 char statstr[160],form[80];
 int ret,l;

 if(mon==NULL)
   return(0);

 if(mon->quiet)
   return(2);

 statstr[0]= 0;
 ret= Pppoem_format_stat(mon,statstr,1|((flag&1)<<2));
 if(ret<=0)
   return(ret);
 Pppoem_show_stat(mon,1);
 if(flag&2) {
   l= 8+1+strlen(statstr)+strlen(mon->dev_name);
   if(l<79) {
     sprintf(form,"%%-8.8s %%s%%%ld.%lds%%s\n",79-l,79-l);
     fprintf(stdout,form,msg,statstr," ",mon->dev_name);
   } else
     fprintf(stdout,"%-8.8s %s\n",msg,statstr);
 } else
   fprintf(stdout,"%-8.8s %s\n",msg,statstr);
 fflush(stdout);
 return(1);
}


int Pppoem_forward(mon,flag)
struct PppoeM *mon;
int flag;
{
 int ret,idx;

 if(mon==NULL)
   return(0);
 if(mon->quiet!=2)
   return(2);

 ret= Pppoem_ring_index(mon,mon->ring_cursor-1,&idx,0);
 if(ret<=0)
   return(ret);
 printf("%ld %.512s %lf %lf\n",mon->is_open,mon->dev_name,
                               mon->ring_times[idx],mon->ring_counts[idx]);
 fflush(stdout);
 return(1);
}


int Pppoem_loop(mon,flag)
struct PppoeM *mon;
int flag;
{
 int ret,was_open;
 unsigned long microsec;
 was_open= -1;

 if(mon==NULL)
   return(0);

 ret= Pppoem_report(mon,"PPPOEM S",1|2);
 if(ret<=0)
   goto ex;
 if(! mon->headlines_on) 
   Pppoem_show_stat(mon,2);                             /* one headline */

 do {
   if(Pppoem_has_lifetime(mon,0)<=0)
 break;
   ret= Pppoem_poll(mon,1);
   if(ret<=0)
     goto ex;
   if(mon->is_open!=was_open) {
     if(mon->is_open==2) {
       ret= Pppoem_report(mon,mon->dev_name,1|2);
     } else if(mon->is_open>0) {
       ret= Pppoem_report(mon,"DSL ON",1|2);
     } else {
       ret= Pppoem_report(mon,"DSL OFF",0);
     }
     if(ret<=0)
       goto ex;
   } else {
     ret= Pppoem_show_stat(mon,0);
     if(ret<=0)
       goto ex;
   }
   Pppoem_forward(mon,0);
   if(!mon->listen_stdin) {
     microsec= mon->delay*1e6;
     Sfile_microsleep(microsec,0);
   }
   if(mon->is_open!=was_open && mon->is_open>0 && mon->headlines_on) 
     Pppoem_show_stat(mon,2);                  /* new headline */
   was_open= mon->is_open;
 } while(1);
 
 ret= 1;
ex:;
 return(ret);
}


int Pppoem_arg_interpreter(mon,argc,argv,flag)
struct PppoeM *mon;
int argc;
char **argv;
int flag;
/*
 bit0= ignore first argument
*/
{
 int ret,i;
 double d;

 for(i=(flag&1);i<argc;i++) {
   if(strcmp(argv[i],"-count")==0 && i+1<argc) {
     i++;
     if(strcmp(argv[i],"received")==0)
       mon->count_what= 1;
     else if(strcmp(argv[i],"sent")==0)
       mon->count_what= 2;
     else
       mon->count_what= 3;
   } else if(strcmp(argv[i],"-dev_name")==0 && i+1<argc) {
     i++;
     strncpy(mon->dev_name,argv[i],sizeof(mon->dev_name)-1);
     mon->dev_name[sizeof(mon->dev_name)-1]= 0;
     mon->dev_automatic= 0;
     if(strcmp(mon->dev_name,"-")==0 || mon->dev_name[0]==0)
        mon->dev_automatic= 1;
   } else if(strcmp(argv[i],"-help")==0 || strcmp(argv[i],"-h")==0) {
     fprintf(stderr,"usage: %s [options]\n\n",argv[0]);
     Pppoem__help_stderr(1);
   } else if(strcmp(argv[i],"-lifetime")==0 && i+1<argc) {
     i++;
     d= -1.0;
     sscanf(argv[i],"%lf",&d);
     if(d<0.0 || d>1.0e9) {
       fprintf(stderr,"-lifetime: not a usable number [1...1e9] : %s\n",
                      argv[i]);
         ret= 0; goto ex;
     }
     if(d<1.0)
       mon->deadline= 0;
     else
       mon->deadline= time(0)+d;
   } else if(strcmp(argv[i],"-listen_stdin")==0 && i+1<argc) {
     i++;
     if(strcmp(argv[i],"on")==0)
       mon->listen_stdin= 1;
     else
       mon->listen_stdin= 0;
   } else if(strcmp(argv[i],"-pretend")==0 && i+1<argc) {
     i++;
     if(strcmp(argv[i],"on")==0)
       mon->is_open= 2;
     else
       mon->is_open= -1;
   } else if(strcmp(argv[i],"-report")==0 && i+1<argc) {
     i++;
     if(strcmp(argv[i],"off")==0)
       mon->quiet= 1;
     else if(strcmp(argv[i],"forward")==0)
       mon->quiet= 2;
     else
       mon->quiet= 0;
   } else if(strcmp(argv[i],"-show")==0 && i+1<argc) {
     int time_spans[3],max_span= -1;
     int j;

     i++;
     for(j=0;j<3;j++)
       time_spans[j]= -1;
     sscanf(argv[i],"%ld:%ld:%ld",time_spans,time_spans+1,time_spans+2);
     for(j=0;j<3;j++) {
       if(time_spans[j]<=0) {
         fprintf(stderr,"-show: no three usable time spans : %s\n",
                        argv[i]);
         ret= 0; goto ex;
       }
       if(time_spans[j]>max_span)
         max_span= time_spans[j];
     }
     ret= Pppoem_new_ring(mon,max_span+1,0);
     if(ret<=0)
       goto ex;
     for(j=0;j<3;j++)
       mon->time_spans[j]= time_spans[j];
   } else if(strcmp(argv[i],"-wait_silently")==0 && i+1<argc) {
     i++;
     if(strcmp(argv[i],"on")==0)
       mon->missing_pppoe_silent= 1;
     else
       mon->missing_pppoe_silent= 0;
   } else {
     fprintf(stderr,"unknown or incomplete option : %s\n",argv[i]);
     ret= 0;
     goto ex;
   }
 }
 ret= 1;
ex:;
 return(ret);
}


